/*
 * graphical representation of the 
 * LABOR MARKET, PRODUCTION FUNCTION, AGGREGATE DEMAND and SUPPLY
 * initial situation and after changes of parameters (updates)
 */
#define RangeOf(x,a,b)  if ( x < a ) x = a; if ( x > b )  x = b
template <class X> X Min(X x, X y) { return (x<=y) ? x : y ;}
template <class X> X Max(X x, X y) { return (x>=y) ? x : y ;}

#include "Equilibrium.h"

#include <QPainter>
#include <QPixmap>
#include <QRect>
#include "ISMarket.h"

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

/********************************************************************/

Equilibrium::Equilibrium( QWidget *parent )
        : QWidget( parent )
{
  FPM_ID = false;      // no Fixed Price Model at the beginning
  Resolution( true ) ; // must be first
  HINTFLAG = true;     // paint gray lines
  OpenEconomy = false; // a closed economy is default
  KeynesCase = 3;
  iMAX1=0.22;
  iMAX2=0.20;
  iMIN2=0.02; 
  k = 1.;
  M = 1.;
  s = 1.;
  alpha=.3;
  Capital_h = 999;       // capital stock K
  Kalpha = pow(double(Capital_h),alpha) ;
  delta= 90./(iMAX2-iMIN2);
  LS0  = 90. + delta*iMIN2;
  P = 1.;
  PKeynes=1.;
  Phint=1.;
  PEq = 1. ;
  wMin = 1.;
  rho = 1. ;
  YDEq = 1. ;
  YSKeynes = 1.;
  ERate = 1.;
  autDemand = 1. ;
  InitialValues( 3 ) ;
  progress=0;
}
void Equilibrium::paintEvent( QPaintEvent * )
{
  QPainter r( this ); 
  DrawAxes( &r );
  paintLaborMarket( &r ) ; 
  paintDemandSide( &r ) ; 
}
void Equilibrium::DrawAxes( QPainter *r )
{
  r->setBrush(QBrush("#fffff0"));  
  r->setPen(Qt::black);
  r->drawRect( 0, 0, 1.999 * CENTER, 1.999 * CENTER );  
  r->setBrush(Qt::NoBrush); 
#if __GNUG__ /* GNU C++ Compiler Linux */
  r->setFont( QFont( "helvetica", (SCALE==0.75) ? 10 : 12, 25 ) ); 
#else        /* C++ for Mickysoft Windows */
  r->setFont( QFont( "arial", (SCALE==0.75) ? 8 : 11, 25 ) ); 
#endif 
  r->drawText( int(SCALE * 165), int(SCALE * 15), "P" );
  r->drawText( int(SCALE * 300), int(SCALE * 155), "Y" );
  r->drawText( int(SCALE * 165), int(SCALE * 310), "N" );
  if (KeynesCase != 2) 
    r->drawText( int(SCALE * 5), int(SCALE * 155), "w/P" ); 
  else 
    r->drawText( int(SCALE * 5), int(SCALE * 155), "w_min/P" ); 
  r->drawText( int(SCALE * 230), int(SCALE * 314), "Y=Y(N,K)" );
  r->drawLine( int(SCALE * 10), CENTER, int(SCALE * 310), CENTER );
  r->drawLine( CENTER, int(SCALE * 10), CENTER, int(SCALE * 310) );
  Ycritical1  = (autDemand - rho*iMIN2)/s;  
  Ycritical2  = (autDemand - rho*iMAX2)/s;  
}
void Equilibrium::paintLaborMarket( QPainter *r )
{
  if (Kalpha==0) Kalpha=.0000001 ;
  QString HBString;
  QPoint p1,p2;
  double betaINV=1./(beta-1.),
         wRealised = 0.,
         LaborD,
         wEq, NEq,
         temp;
  /*
   *  number of points to be shown 
   */
  int temp1 = (SCALE < 1.) ? 113 : 150; 
  /*
   *   equilibrium values (supply side)
   *   wEq = nominal wage rate
   */ 
  if (P==0) P=.000000001 ;
  wEq  = pow( pow( gamma, beta - 1.) * beta * Kalpha, 1./(2.-beta) ) * P ;
  NEq  = wEq / P * gamma ;
  YSEq = Kalpha * pow ( NEq, beta);
  /*
   *  Labor demand
   */
  QPolygon NDemand( temp1 ); 
  for (int i=0; i<temp1; i++ ) {     // 0 < WageRate < 3.
    double w_real = (i+1) / wscaling;
    LaborD = Min( int(Nscaling * pow( w_real/(beta*Kalpha), betaINV )), CENTER+1);
    NDemand.setPoint(i, - (i+1), int(LaborD));
  }
  DrawCurve( r, NDemand, Qt::black ) ; 
  /*
   *  Labor supply
   */
  if (gamma * 3. <= 3000.) {
    p1 = QPoint( 0, CENTER + int(Nscaling * gamma * 3. * 160./150.) );
    r->drawLine( Origin, p1 ); }
  else {
    p1 = QPoint( -int(wscaling * 3000 * 160./150. / gamma), CENTER ) + Origin ;
    r->drawLine( Origin, p1 ); }
  HBString.sprintf( " N*  =%7.1f ", NEq);
  r->drawText( int(SCALE * 25), int(SCALE * 40), HBString );
  /*
   *  production function
   */
  QPolygon ProdFct( temp1 ); 
  for (int i=0; i<temp1; i++ ) {  // 0 < Y < 2000 
    LaborD = pow( (i/Yscaling)/Kalpha, 1./beta ) ;
    if (Nscaling * LaborD > 161. )
       ProdFct.setPoint(i, i, 161);
    else    
       ProdFct.setPoint(i, i, int(Nscaling * LaborD));
  }
  DrawCurve( r, ProdFct, Qt::black ) ;
  if (KeynesCase==2) { 
    wRealised = Max( wMin, wEq );
    /*
     *   gray hint lines
     *   w_min/P -- NEq -- YEq -- PEq
     */
    if (HINTFLAG==true) {
      r->setPen(Qt::lightGray);
      if ( wEq <= wMin ) { 
	PEq = pow( pow( gamma, beta - 1.) * beta * Kalpha, 1./(beta-2.)) 
	  * wRealised ;
      }
      else {
	PEq = P;
      }
      NEq = pow( (wRealised / PEq) / (beta * Kalpha), betaINV);
    }
    /*
     *  YS-curve (magenta) 
     */
    QPolygon YKeynes( temp1 ); 
    for (int i=0; i<temp1; i++ ) {
      double w_real = (i+10)/wscaling ;
      if (w_real<=wRealised/PEq) {
	YKeynes.setPoint(i, int(Yscaling * YSEq), int(-Pscaling * PEq));
      }
      else {
	PKeynes = wRealised / w_real;
	NKeynes = pow( w_real / (beta * Kalpha) , betaINV );
	YSKeynes = pow( NKeynes, beta ) * Kalpha ;
	YKeynes.setPoint(i, int(Yscaling * YSKeynes),-int(Pscaling * PKeynes));
      }
    }
    DrawCurve( r, YKeynes, Qt::magenta ) ;
    r->setPen(Qt::magenta);
    p1 = QPoint( int(Yscaling * YSEq), -int(Pscaling * PEq) ) + Origin ;
    p2 = QPoint( CENTER + int(Yscaling * YSEq), 0 );
    r->drawLine( p1, p2 );
    /*
     *  aggregate supply based on demand side price level
     */
    if (P<PEq) {
      YSEqatPD = Kalpha * pow( (wMin/P)/(beta * Kalpha), beta/(beta-1.) ) ;
      NSatPD = gamma * wMin / P ;
    }
    else { 
      YSEqatPD = YSEq ; 
      NSatPD = gamma * wMin / PEq ;
    }
    HBString.sprintf( " YS* =%7.1f ", YSEqatPD);
    r->drawText( int(SCALE * 25), int(SCALE * 60), HBString );
    r->setPen(Qt::black);
    if (YSEqatPD<=YDEq) {
      NDatPD = pow(YSEqatPD/Kalpha, 1./beta);
      Phint = P ;
    }
    else { 
      NDatPD = pow(YDEq/Kalpha, 1./beta);
      Phint = wMin/(beta * Kalpha) * pow(YDEq/Kalpha,(1.-beta)/beta) ;
      NSatPD = gamma * wMin / Phint ;
    }
    if (YSEqatPD<=YDEq) {
      if ( wEq <= wMin ) { 
	/*
	 * the reason for unemployment is w_min>wEq
	 *
	 * black paranoia hyperbola ;-)
	 * P = w_min / (w_min/P) 
	 */
	r->setPen(Qt::black);
	QPolygon b( temp1-10 );
	for (int i=0; i<temp1-10; i++ ) {
	  double w_real = (i+10)/wscaling ;
	  temp = Min( CENTER + 1 , int(Pscaling * wMin / w_real) );
	  b.setPoint(i, -(i+10), -int(temp) ); 
	}
	DrawCurve( r, b, Qt::black ) ; 
	r->setPen(Qt::red);
	HBString.sprintf( " w*  =%7.4f ", wEq); 
	r->drawText( int(SCALE * 25), int(SCALE * 20), HBString );
	HBString.sprintf( " NU  =%7.1f ", NSatPD-NDatPD);
	/*
	 *  show unemployment 
	 */
	p1 = QPoint(-int(wscaling * wMin / P), int(Nscaling * NSatPD)) + Origin;
	p2 = QPoint(-int(wscaling * wMin / P), int(Nscaling * NDatPD)) + Origin;
	r->drawLine(p1,p2);
	if (HINTFLAG==true) {
	  r->setPen(Qt::lightGray);
	  p1 = QPoint(-int(wscaling * wRealised / PEq),-int(Pscaling *PEq)) + Origin;
	  p2 = QPoint( int(Yscaling * YSEq), int(Nscaling * NEq))           + Origin;
	  r->drawRect(QRect(p1,p2));

	  p1 = QPoint(-int(wscaling * wMin / P),-int(Pscaling * P))      + Origin ;
	  p2 = QPoint( int(Yscaling * YSEqatPD), int(Nscaling * NDatPD)) + Origin ;
	  r->drawRect(QRect(p1,p2));
	}
      }
      else {
	r->setPen(Qt::red);
	HBString.sprintf( " w* > w_min ");
      }
    }
    else {
      /*
       * the reason for unemployment is now YD<YS (not w_min>wEq )
       */
      NDatPD = pow(YDEq/Kalpha, 1./beta);
      wEq = beta * Kalpha * pow(NDatPD,beta-1.) * P ;
      r->setPen(Qt::green);
      HBString.sprintf( " w*  =%7.4f ", wEq); 
      /*
       *  tiny green bar showing real wage rate 
       */
      r->drawText( int(SCALE * 25), int(SCALE * 20), HBString );
      p1 = QPoint(-int(wscaling * wEq/P), -4 ) + Origin ;
      p2 = QPoint(-int(wscaling * wEq/P),  4 ) + Origin ;
      r->drawLine( p1, p2 );
                   
      /*
       *  show unemployment 
       */
      r->setPen(Qt::red);
      HBString.sprintf( " NU  =%7.1f ", NSatPD-NDatPD);
      p1 = QPoint(-int(wscaling * wEq/P), int(Nscaling * NSatPD)) + Origin ;
      p2 = QPoint(-int(wscaling * wEq/P), int(Nscaling * NDatPD)) + Origin ;
      r->drawLine( p1, p2 );

      if (HINTFLAG==true) {
	r->setPen(Qt::lightGray);
	p1 = QPoint(-int(wscaling * wRealised / PEq), 0 ) + Origin ;
	p2 = QPoint(-int(wscaling * wRealised / PEq), int(Nscaling * NEq) ) + Origin ;
	r->drawLine( p1, p2 );
	p1 = QPoint( int(Yscaling * YSEq), int(Nscaling * NEq) ) + Origin ;
	r->drawLine( p2, p1 );
	p2 = QPoint( int(Yscaling * YSEq),-int(Pscaling * PEq) ) + Origin ;
	r->drawLine( p1, p2 );
	r->setPen(Qt::green);
	p1 = QPoint( int(Yscaling * YDEq),-int(Pscaling * P) ) + Origin ;
	p2 = QPoint( int(Yscaling * YDEq), int(Nscaling * NDatPD) ) + Origin ;
	r->drawLine( p1, p2 );
	p1 = QPoint(-int(wscaling * wEq/P), int(Nscaling * NDatPD)) + Origin ;
	r->drawLine( p2, p1 );
      }
    }
    // the tiny red bar showing P
    r->setPen(Qt::red);
    r->drawText( int(SCALE * 25), int(SCALE * 120), HBString );
    p1 = QPoint( int(SCALE * 4), -int(Pscaling * P) ) + Origin ;
    p2 = QPoint(-int(SCALE * 4), -int(Pscaling * P)) + Origin ;
    r->drawLine( p1, p2 );
  }
  /*
   *
   * i.e. KeynesCase != 2 
   *
   */
  else {
    if ( YDEq<YSEq ) { 
      /*
       *  show unemployment 
       */
      r->setPen(Qt::red);
      NDatPD = pow(YDEq/Kalpha, 1./beta);
      temp = beta * Kalpha * pow(NDatPD,beta-1.) ;
      NSatPD = gamma * temp ;
      HBString.sprintf( " NU  =%7.1f ", NSatPD-NDatPD);
      r->drawText( int(SCALE * 25), int(SCALE * 120), HBString );
      p1 = QPoint(-int(wscaling * temp), int(Nscaling * NSatPD) ) + Origin ;
      p2 = QPoint(-int(wscaling * temp), int(Nscaling * NDatPD) ) + Origin ;
      r->drawLine( p1, p2 );
      if (HINTFLAG==true) {
	/*
	 *  the gray lines show
	 *  YS>YD --> YD=Y(KD,ND) --> ND<NS (unemployment) 
	 */ 
	temp = Kalpha * pow(NDatPD,beta) ;
	r->setPen(Qt::gray);
	p1 = QPoint( int(Yscaling * temp), int(Nscaling * NDatPD) ) + Origin ;
	r->drawLine( p2, p1 );
	p2 = QPoint( int(Yscaling * temp),-int(Pscaling * P) ) + Origin ;
	r->drawLine( p1, p2 );
	p1 = QPoint( int(Yscaling * YSEq),-int(Pscaling * P) ) + Origin ;
	r->drawLine( p2, p1 );
      }
      r->setPen(Qt::black);
    }
    /*
     *  YS-curve (magenta) 
     */
    r->setPen(Qt::magenta);
    if (FPM_ID==true) {
      p1 = QPoint( 0,-int(Pscaling * P) ) + Origin ;
      p2 = QPoint( int(Yscaling * YSEq),-int(Pscaling * P) ) + Origin ;
      r->drawLine( p1 , p2 );
      p1 = QPoint(  CENTER + int(Yscaling * YSEq), 0 );
      r->drawLine( p2 , p1 );
    }
    else { 
      r->drawLine(CENTER + int(Yscaling * YSEq), CENTER, 
		  CENTER + int(Yscaling * YSEq), 0 );
    }
    YSEqatPD = YSEq; // needed for adjustment of P
    HBString.sprintf( " YS* =%7.1f ", YSEq);
    r->drawText( int(SCALE * 25), int(SCALE * 60), HBString );
    if (HINTFLAG==true) {
      /*
       *   vertical line at w/P (equilibrium)
       *   horizontal line at NEq 
       *   vertical line at YEq (supply side) 
       */ 
      r->setPen(Qt::lightGray);
      p1 = QPoint(-int(wscaling * wEq / P), 0 ) + Origin ;
      p2 = QPoint(-int(wscaling * wEq / P), int(Nscaling * NEq) ) + Origin ;
      r->drawLine( p1, p2 );
      p1 = QPoint( int(Yscaling * YSEq), int(Nscaling * NEq) ) + Origin ;
      r->drawLine( p2, p1 );
      if (FPM_ID==true) {
        p2 = QPoint( int(Yscaling * YSEq),-int(Pscaling * P) ) + Origin ;
	r->drawLine( p1, p2 ); }
      else {
        p2 = QPoint( int(Yscaling * YSEq), 0 ) + Origin ;
	r->drawLine( p1, p2 ); }
    }
    r->setPen(Qt::black);
    HBString.sprintf( " w*  =%7.4f ", wEq); 
    r->drawText( int(SCALE * 25), int(SCALE * 20), HBString );
  }
}
void Equilibrium::paintDemandSide( QPainter *r )
{
  QString HBString;
  QPoint p1,p2;
  double PDemand, temp ;
  /*
   *  P=P(Y) -> aggregate demand curve YD
   * 
   *  classical region
   *  liquidity trap 
   *  investment trap 
   *
   *  YDEq, YatiMAXonLM, YatiMINonLM, YatiMAXonIS are known from IS-LM.cpp
   */
  if (YatiMAXonIS>YatiMAXonLM) { // no intersection for IS-LM
    r->setPen(Qt::red);
    r->drawPixmap(int(SCALE * 320 - 38), int(SCALE * 5), 
		  QPixmap("icons/bug.png"));
    r->setPen(Qt::black);
    return; 
  }
  if (HINTFLAG==true) {
    /*
     *   vertical line at YDEq (demand side) 
     *   horizontal line at P
     */
    r->setPen(Qt::lightGray);
    if (Yscaling * YDEq<= CENTER ) {     
      p2 = QPoint( CENTER + int(Yscaling * YDEq), 0 );
      p1 = QPoint( CENTER + int(Yscaling * YDEq), CENTER - int(Pscaling * P));
      r->drawLine( p1, p2 );
    }
    else {
      p1 = QPoint( 2*CENTER, CENTER - int(Pscaling * P));
    }
    if (FPM_ID==true) 
      p2 = QPoint( CENTER + int(Yscaling * YSEq), CENTER - int(Pscaling * P));
    else 
      p2 = QPoint( CENTER, CENTER - int(Pscaling * P));
    r->drawLine( p1, p2 );  // ??? richtiges p1 ?
  }
  /*
   *  number of points to be shown 
   */
  int temp1 = (SCALE < 1.) ? 113 : 150 ;  
  /*
   *  aggregate demand YD
   */
  QPolygon a( temp1 );
  if ( OpenEconomy == true ) {
    double y0,y1,y2,f0,f1,f2,CAB0,CAB1,CAB2;
    /*
     *   compute smallest Y (both the IS- and the LM-curve depend on P!)
     */
    bool REPEAT = true;
    y0 = .001 ;
    CAB0 = ISMarket->BoPA(y0,4500.,(M/k/y0),2.,ERate);
    f0 = s * y0 - CAB0 - autDemand + rho * iMAX1 ;
    y1 = YatiMAXonLM ;
    CAB1 = ISMarket->BoPA(y1,4500.,(M/k/y1),2.,ERate);
    f1 = s * y1 - CAB1 - autDemand + rho * iMAX1 ;
    if (f0 * f1 < 0.) {
      while (REPEAT == true) {
	y2 = (y0 * f1 - y1 * f0) / (f1 - f0);
	CAB2 = ISMarket->BoPA(y2,4500.,(M/k/y2),2.,ERate);
	f2 = s * y2 - CAB2 - autDemand + rho * iMAX1 ;
	if ( fabs(f2) < 0.2 ) {
	  YatiMAXonIS = y2 ; 
	  REPEAT = false ;
	} 
	else {
	  if (f2 * f0 < 0. ) 
	    { y1 = y2; f1 = f2; }
	  else 
	    { y0 = y2; f0 = f2; }
	}
      }
    }
    else {     
      /*
       *  f1>>0 because of  y1>>0 (dominates CAB1<<0)
       *  f0>0  because of  CAB0<<0 (at y=0) 
       */
      if (f0>0) { YatiMAXonIS = 0. ; }
    }
    /*
     *   aggregate demand Y=Y(P) 
     */
    for (int j=0; j<temp1; j++ ) {
      PDemand = j / Pscaling ;
      if (PDemand==0) PDemand = 0.0000001 ; 
      temp = -autDemand+(rho/delta)*(LS0-M/PDemand) ;
      REPEAT = true ;
      y0 = 0. ;
      CAB0 = ISMarket->BoPA(y0,4500.,PDemand,2.,ERate);
      f0 = - CAB0 + temp ;
      y1 = Max( 2150.,  YatiMAXonLM ) ;
      CAB1 = ISMarket->BoPA(y1,4500.,PDemand,2.,ERate);
      f1 = (s+rho*k/delta) * y1 - CAB1 + temp ;

      if (f0 * f1 < 0.)	{ 
	while (REPEAT == true) { 
	  y2 = (y0 * f1 - y1 * f0) / (f1 - f0);
	  CAB2 = ISMarket->BoPA(y2,4500.,PDemand,2.,ERate);
	  f2 = (s+rho*k/delta) * y2 - CAB2 + temp ;
	
	  if ( fabs(f2) < 0.2 ) {
	    // classical region
	    y2 = Min( y2, (M/PDemand)/k );
	    double i_tmp = (k * y2+LS0-M/PDemand)/delta ;
	    double P_tmp = j;
	    // liquidity trap
	    if ( i_tmp < iMIN2) 
	      y2 = ISMarket->Y(iMIN2,4500.,PDemand,2.,ERate,s,autDemand,rho);
	    // investment trap
	    if ( i_tmp > iMAX1 ) {
	      if (iMAX1<iMAX2) 
		y2 = ISMarket->Y(iMAX1,4500.,PDemand,2.,ERate,s,autDemand,rho);
	      else P_tmp = CENTER;
	    }
	    a.setPoint(j, int(Yscaling * y2), int(-P_tmp) ); 
	    REPEAT = false;
	  } 
	  else {
	    if (f2 * f0 < 0. ) 
	      { y1 = y2; f1 = f2; }
	    else 
	      { y0 = y2; f0 = f2; }
	  }
	}
      }
      else { 
	if (f0<0) {
          temp = CENTER ;
	  if (iMAX1>iMAX2) { 
	    if ( k * y1+LS0-M/PDemand < delta * iMIN2 ) { // liq trap ?
	      y2 = Min( 2300., ISMarket->Y(iMIN2,4500.,PDemand,2.,ERate,s,autDemand,rho) );
	      if (y2>0.1) { temp = Yscaling * y2; }
	    }
	  }
	  a.setPoint(j, int(temp), -j); 
	}
	else { // f0>0
	  if (iMAX1<iMAX2) { 
	    if ( LS0-M/PDemand > delta * iMAX1 ) { // inv' trap ?
	      y2 = Min(2300., ISMarket->Y(iMAX1,4500.,PDemand,2.,ERate,s,autDemand,rho) );
	      a.setPoint(j, int(Yscaling * y2), -j); 
	    }
	    else a.setPoint(j, 0, -CENTER);
	  }
	  else a.setPoint(j, CENTER, -j);
	}
      }
    }
  }
  else { 
    /*
     *  aggregate demand YD (closed economy)
     */
    for (int i=0; i<temp1; i++ ) {
	double YDemand = i/Yscaling ;
	if ( YDemand>Ycritical2 ) { // LS(i)>0
	   PDemand = M/( LS0 + k * YDemand 
		   + (delta/rho) * ( s * YDemand - autDemand )) ;
	}
	else { // LS(i)=0
           if (YDemand > 0)
 	      PDemand = M/( k*YDemand ) ;
  	   else 
	      PDemand = 10.;
	}
	temp = Pscaling * PDemand ;
	if ( (PDemand<0) || (PDemand>3.34) || (YDemand<YatiMAXonIS) ) 
	  { temp = CENTER + 1 ; }  // invisible point
	else { 
	  if ( YDemand>Ycritical1 ) temp = 0 ;
	}
	a.setPoint(i, i, int(-temp) );
    }
  }
  DrawCurve( r, a, Qt::blue ) ;
  r->setPen(Qt::blue);
  HBString.sprintf( " YD* =%7.1f ", YDEq);
  r->drawText( int(SCALE * 25), int(SCALE * 80), HBString );
  if (KeynesCase==2) { r->setPen(Qt::blue); }
  HBString.sprintf( " PD  =%7.3f ", P);
  r->drawText( int(SCALE * 25), int(SCALE * 100), HBString );
  r->setPen(Qt::black);
}
void Equilibrium::UpdateEquilibrium(double *VarList, 
				    bool OpenEconomyNew) 
{
  k    =*VarList;
  LS0  =*(VarList+1);
  delta=*(VarList+2);
  M    =*(VarList+3);
  P    =*(VarList+4);
  autDemand = *(VarList+5) ;    // = Caut + I0 + G 
  s    =*(VarList+6);
  rho  =*(VarList+7);
  iMAX1=*(VarList+8);
  iMAX2=*(VarList+9);
  ERate=*(VarList+10);
  YDEq =*(VarList+11);
  YatiMINonLM=*(VarList+12);
  YatiMAXonLM=*(VarList+13);
  YatiMAXonIS=*(VarList+14);
  OpenEconomy=OpenEconomyNew;
  update();
}
void Equilibrium::UpdateLaborMarket(double CapitalNew,
				    double wminNew)
{
  Kalpha = pow( CapitalNew, alpha );
  if ( KeynesCase==2 ) wMin = wminNew; 
  update();
}
void Equilibrium::HintLines( bool OnOffNew )
{
  HINTFLAG=OnOffNew;
  update();
}
void Equilibrium::Resolution( bool LowResolution )
{
  SCALE = (LowResolution == true) ? 0.75 : 1.0 ;
  CENTER   = int(SCALE * 160) ;
  Origin   = QPoint( CENTER, CENTER );
  Yscaling = SCALE * 150. / 2000. ;
  Nscaling = SCALE * 150. / 3000. ;
  Pscaling = SCALE * 150. / 3.333333 ;
  wscaling = SCALE * 150. / 3. ;
  update();
}
void Equilibrium::Adjustment( int ID )
{
  if (ID==0) { 
    emit updateProgressBar( 0 );
  }
  else {
    progress=0; 
    if ((FPM_ID==false) && (YatiMAXonIS<YatiMAXonLM)) {
      double deltaP = 1.;
      double P_LOCAL = P ;
      QString HBString;
      if (fabs(YDEq-YSEqatPD)<100.) { deltaP = 0.1; }
      if (fabs(YDEq-YSEqatPD)< 40.) { deltaP = 0.01; }
      while ((fabs(YDEq-YSEqatPD)>0.1)
	     && (P<3.34) && (P>0.101)
	     && (deltaP>0.00012)) 
	// smaller deltaP are not recognized by set_P
	{ 
	  deltaP = deltaP * 0.1; 
	  if (deltaP<0.00012) return;
	  if (YDEq>YSEqatPD) {
	    while ((YDEq>YSEqatPD) && (P<3.34)) 
	      {
		emit updateProgressBar( ++progress );
		P_LOCAL = P_LOCAL + deltaP;
		/*
		  Caution: if deltaP is too small, P seems to be constant for
		  the system.
		 */
		set_P( int(P_LOCAL * 1000.) );
		Sleeper::msleep(250);
	      }
	  }
	  else {
	    while ((YDEq<YSEqatPD) && (P>0.101))
	      {
		emit updateProgressBar( ++progress );
		P_LOCAL = P_LOCAL - deltaP;
		set_P( int(P_LOCAL * 1000.) );
		Sleeper::msleep(250);
	      }
	  }
       }
    }
  }
}
void Equilibrium::AdjustPriceLevel( )
{
  Adjustment( 0 );
  Adjustment( 1 );
  emit updateProgressBar( 0 );
}
void Equilibrium::mousePressEvent( QMouseEvent *e )
{
  QString HBString, tmpX, tmpY;
  if ( ( e->button() == Qt::LeftButton ) || ( e->button() == Qt::RightButton ) ) {
    HBString.sprintf( " Mausposition = (%d,%d), ", e->x(),e->y() );
    if ( e->x() > CENTER ) 
      tmpX.sprintf( " Y = %7.1f, ", ( e->x() - CENTER ) / Yscaling );
    else 
      tmpX.sprintf( " w/P = %7.5f, ", ( CENTER - e->x() ) / wscaling );
    if ( e->y() > CENTER ) 
      tmpY.sprintf( " N = %7.1f", ( e->y() - CENTER ) / Nscaling );
    else 
      tmpY.sprintf( " P = %7.5f", ( CENTER - e->y() ) / Pscaling );
    HBString += tmpX ;
    HBString += tmpY ;
    emit explain( HBString );
  }
}
void Equilibrium::mouseMoveEvent( QMouseEvent *e ) 
{
  QString HBString, tmp;
  if (Qt::RightButton) {
    HBString.sprintf(" A parameter has changed: ");
    if ( (e->x() > CENTER) && (e->y() > CENTER) ) {
      Kalpha = ( ( e->x() - CENTER ) / Yscaling ) 
  	     / pow( ( ( e->y() - CENTER ) / Nscaling ), beta ) ;
      tmp.sprintf(" Capital = %7.2f", pow(Kalpha,1./alpha) );
      HBString += tmp ;
      set_K( int(pow(Kalpha,1./alpha)) );
      emit explain( HBString );
      update();
    }
    if ( (e->x() < CENTER) && (e->y() > CENTER) ) {
      gamma = ( ( e->y() - CENTER ) / Nscaling )
	    / ( ( CENTER - e->x() ) / wscaling );
      tmp.sprintf(" gamma = %7.2f", gamma );
      HBString += tmp ;
      emit explain( HBString );
      update();
    }
  }
}
void Equilibrium::DrawCurve ( QPainter *r, 
			      QPolygon Z, 
			      QColor color )
{
  r->setPen(color);
  Z.translate ( CENTER, CENTER );
  r->drawPolyline( Z ); 
  r->setPen(Qt::black);
}
void Equilibrium::InitialValues ( int KeynesNew ) 
{
  P_h = 1000; 
  w_h = 0; 
  alpha_h=3333;
  alpha= alpha_h/10000.;
  beta = 1.-alpha;
  gamma= 15.*3000. / (14.*3.);
  KeynesCase = KeynesNew ;
  switch (KeynesCase) {
  case 0: // liquidity trap
    K_h = 3900; 
    break ;
  case 1: // investment trap 
    P_h  = ( OpenEconomy==true ) ? 1500 : 2100 ; 
    K_h =  500;
    break ; 
  case 2: // wage rate rigidity
    K_h = 1000; 
    w_h = 1000;
    break ;
  case 3: // usual case
    K_h = 1000; 
    break ;
  }
  set_w(0);
  set_alpha(10);
  set_P(10);                // LCD reagiert sonst nicht !?
  set_K(10);
  set_w(w_h);
  set_P(P_h);
  set_K(K_h);
  set_alpha(alpha_h);
  update();
}
void Equilibrium::Growth () 
{
  QString HBString, tmp;
  double time ;
  double GrowthRate = 0.011 ;
  double DepreciationRate = 0.24 ;
  double gamma0 = 10.*3000. / (14.*3.) ;
  double K0 = 1000. ;
  double CapitalStock = K0 ;
  double Y0, P0, w0, i0, M0, CautNew, Iaut, GNew ;
  P0 = 1.5;
  set_P( int(P0 * 1000.) );
  Kalpha = pow( CapitalStock, alpha) ;
  w0 = pow( pow( gamma0, beta - 1.) * beta * Kalpha, 1./(2.-beta) ) * P0 ;
  Y0 = Kalpha * pow(gamma0 * w0/P0, beta);
  i0 = - ( s * Y0 - autDemand ) / rho ;
  M0 = P0 * ( LS0 - delta * i0 + k * Y0 ) ;
  emit MChanged( int(M0 * 10.) );
  repaint();
  HBString.sprintf("i, P, w are const.");
  tmp.sprintf(", Y(0) = %6.1f", Y0); HBString += tmp;
  tmp.sprintf(", M(0) = %5.1f", M0); HBString += tmp;
  tmp.sprintf(", Caut(0) = 99.2");   HBString += tmp;
  emit explain( HBString );
  for (time=0; time<101; time++) {
    double timeFactor = exp( GrowthRate * time );
    gamma = gamma0 * timeFactor ;
    CapitalStock = K0 * timeFactor ;
    set_K( int(CapitalStock) );
    Kalpha = pow( CapitalStock, alpha) ;
    M = P0 * ( LS0 - delta * i0 + k * Y0 * timeFactor ) ;
    emit MChanged( int(M * 10.) );
    GNew = 50. * timeFactor ;
    //CautNew = ( s * Y0 -(GRate+DRate) * K0 -50. ) * timeFactor;
    CautNew = 99.238024 * timeFactor;
    Iaut = (GrowthRate+DepreciationRate) * CapitalStock + rho * i0 ;
    emit IautChanged( CautNew, GNew, Iaut );  // Iaut includes I0 !!
    Sleeper::msleep(250);
    repaint();
    if (time==0) {
      QMessageBox::warning(this,tr("Growth"), 
			   tr("MESSAGE Growth"),
			   QMessageBox::Ok);
    }
  }
}
void Equilibrium::set_K( int tmp ) 
{
  RangeOf(tmp,1,10000);
  if ( Capital_h == tmp )  return; // don't play ping-pong with LCD
  Capital_h = tmp ;
  emit KChanged( tmp );            // signal to LCD
  Kalpha = pow(double(tmp), alpha );
  update();
}
void Equilibrium::set_P( int tmp ) 
{
  RangeOf(tmp,1,10000);
  if ( P == tmp/1000. )  return; // don't play ping-pong with LCD
  P = tmp / 1000.;
  emit PChanged( tmp );          // signal to LCD
  update();
}
void Equilibrium::set_w( int tmp ) 
{
  RangeOf(tmp,0,10000);
  if ( wMin == tmp/1000. )  return; // don't play ping-pong with LCD
  wMin = tmp / 1000.; 
  if ( (KeynesCase!=2) && (wMin!=0)) {
    KeynesCase=2;
    emit wminClicked();
    /*    helpmessage = new QMessageBox(QString("Wage Rate Rigidities"),  
				  QString("MESSAGE WRR"),
                  QMessageBox::Warning, QMessageBox::Ok, 0, 0, 0, 0, false );
    helpmessage->setIconPixmap( QPixmap( HBWarning ));
    helpmessage->show();
    */
    QMessageBox::warning(this,tr("Wage Rate Rigidities"),
			 tr("MESSAGE WRR"),
			 QMessageBox::Ok);
  }
  emit wChanged( tmp );            // signal to LCD
  update();
}
void Equilibrium::set_alpha( int tmp ) 
{
  RangeOf(tmp,60,9950);
  if ( alpha == tmp/10000. )  return; // don't play ping-pong with LCD
  alpha = tmp / 10000.; 
  beta = 1.- alpha ; 
  Kalpha = pow(Capital_h,alpha) ;
  emit alphaChanged( tmp );           // signal to LCD 
  update();
}
void Equilibrium::FPMChanged( int ID )
{
  FPM_ID = ID;
  update();
}
